/*      Copyright (c) 2015 Regents of the University of California
 *
 *      Author: Brian Demsky <bdemsky@uci.edu>
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      version 2 as published by the Free Software Foundation.
 */

#ifndef HASHSET_H
#define HASHSET_H
#include "hashtable.h"

template<typename _Key, typename _KeyInt, int _Shift, unsigned int (*hash_function)(_Key), bool (*equals)(_Key, _Key)>
class Hashset;

template<typename _Key, typename _Val, typename _KeyInt = uintptr_t, int _Shift = 0, unsigned int (*hash_function)(_Key) = defaultHashFunction<_Key, _Shift, _KeyInt>, bool (*equals)(_Key, _Key) = defaultEquals<_Key> >
class SetIterator {
public:
	SetIterator(Hashlistnode<_Key, _Val> *_curr, Hashtable <_Key, _Val, _KeyInt, _Shift, hash_function, equals> *_table) :
		curr(_curr),
		table(_table)
	{
	}

	/** Override: new operator */
	void *operator new(size_t size) {
		return ourmalloc(size);
	}

	/** Override: delete operator */
	void operator delete(void *p, size_t size) {
		ourfree(p);
	}

	/** Override: new[] operator */
	void *operator new[](size_t size) {
		return ourmalloc(size);
	}

	/** Override: delete[] operator */
	void operator delete[](void *p, size_t size) {
		ourfree(p);
	}

	bool hasNext() {
		return curr != NULL;
	}

	_Key next() {
		_Key k = curr->key;
		last = curr;
		curr = curr->next;
		return k;
	}

	_Key currKey() {
		return last->key;
	}

	void remove() {
		_Key k = last->key;
		table->remove(k);
	}

private:
	Hashlistnode<_Key,_Val> *curr;
	Hashlistnode<_Key, _Val> *last;
	Hashtable <_Key, _Val, _KeyInt, _Shift, hash_function, equals> *table;
};

template<typename _Key, typename _KeyInt = uintptr_t, int _Shift = 0, unsigned int (*hash_function)(_Key) = defaultHashFunction<_Key, _Shift, _KeyInt>, bool (*equals)(_Key, _Key) = defaultEquals<_Key> >
class Hashset {
public:
	Hashset(unsigned int initialcapacity = 16, double factor = 0.5) :
		table(new Hashtable<_Key, _Key, _KeyInt, _Shift, hash_function, equals>(initialcapacity, factor))
	{
	}

	/** @brief Hashset destructor */
	~Hashset() {
		delete table;
	}

	Hashset<_Key, _KeyInt, _Shift, hash_function, equals> *copy() {
		Hashset<_Key, _KeyInt, _Shift, hash_function, equals> *copy = new Hashset<_Key, _KeyInt, _Shift, hash_function, equals>(table->getCapacity(), table->getLoadFactor());
		SetIterator<_Key, _Key, _KeyInt, _Shift, hash_function, equals> *it = iterator();
		while (it->hasNext())
			copy->add(it->next());
		delete it;
		return copy;
	}

	void clear() {
		table->clear();
	}

	/** @brief Adds a new key to the hashset.  Returns false if the key
	 *  is already present. */

	void addAll(Hashset<_Key, _KeyInt, _Shift, hash_function, equals> *table) {
		SetIterator<_Key, _Key, _KeyInt, _Shift, hash_function, equals> *it = iterator();
		while (it->hasNext())
			add(it->next());
		delete it;
	}

	/** @brief Adds a new key to the hashset.  Returns false if the key
	 *  is already present. */

	bool add(_Key key) {
		if (!table->contains(key)) {
			table->put(key, key);
			return true;
		} else
			return false;
	}

	/** @brief Gets the original key corresponding to this one from the
	 *  hashset.  Returns NULL if not present. */

	_Key get(_Key key) {
		return table->get(key);
	}

	_Key getFirstKey() {
		return table->list->key;
	}

	bool contains(_Key key) {
		return table->contains(key);
	}

	bool remove(_Key key) {
		if (!table->contains(key))
			return false;
		table->remove(key);
		return true;
	}

	unsigned int size() const {
		return table->size();
	}

	bool isEmpty() const {
		return size() == 0;
	}

	SetIterator<_Key, _Key, _KeyInt, _Shift, hash_function, equals> *iterator() {
		return new SetIterator<_Key, _Key, _KeyInt, _Shift, hash_function, equals>(table->list, table);
	}

	/** Override: new operator */
	void *operator new(size_t size) {
		return ourmalloc(size);
	}

	/** Override: delete operator */
	void operator delete(void *p, size_t size) {
		ourfree(p);
	}

	/** Override: new[] operator */
	void *operator new[](size_t size) {
		return ourmalloc(size);
	}

	/** Override: delete[] operator */
	void operator delete[](void *p, size_t size) {
		ourfree(p);
	}
private:
	Hashtable<_Key, _Key, _KeyInt, _Shift, hash_function, equals> *table;
};

template<typename _Key, typename _Val, typename _KeyInt, int _Shift, unsigned int (*hash_function)(_Key), bool (*equals)(_Key, _Key)>
SetIterator<_Key, _Val,_KeyInt, _Shift, hash_function, equals> *getKeyIterator(Hashtable<_Key,_Val,_KeyInt,_Shift,hash_function,equals> *table) {
	return new SetIterator<_Key, _Val, _KeyInt, _Shift, hash_function, equals>(table->list, table);
}
#endif
